VERSION = (
    0,
    4,
    11,
)

__title__ = "tsd"
__description__ = "Search and download of Sentinel, Landsat and Planet crops."
__url__ = 'https://github.com/cmla/tsd'
__version__ = ".".join(map(str, VERSION))
__author__ = "Carlo de Franchis"
__author_email__ = 'carlo.de-franchis@ens-cachan.fr'
